package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.model.AnsibleSpec;
import be.iminds.ilabt.jfed.espec.model.ExecuteSpec;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.highlevel.DesiredStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Provider;
import javax.inject.Singleton;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * User: twalcari
 * Date: 11/27/2015
 * Time: 17:12
 */
@Singleton
public class JobStateFactory {
    @Nonnull private final Provider<HighLevelTaskFactory> hltfProvider;
    @Nonnull private final Provider<AggregateManagerWrapperFactory> aggregateManagerWrapperFactoryProvider;
    @Nonnull private final GeniUserProvider geniUserProvider;
    @Nonnull private final Provider<SfaModel> sfaModelProvider;
    @Nonnull private final Provider<ExperimentPartControllerManager> epcManagerProvider;
    @Nonnull private final ProxySocketFactoryProvider proxySocketFactoryProvider;

    @Inject
    public JobStateFactory(@Nonnull Provider<HighLevelTaskFactory> hltfProvider,
                           @Nonnull Provider<AggregateManagerWrapperFactory> aggregateManagerWrapperFactoryProvider,
                           @Nonnull GeniUserProvider geniUserProvider,
                           @Nonnull Provider<SfaModel> sfaModelProvider,
                           @Nonnull Provider<ExperimentPartControllerManager> epcManagerProvider,
                           @Nonnull ProxySocketFactoryProvider proxySocketFactoryProvider) {
        this.hltfProvider = hltfProvider;
        this.aggregateManagerWrapperFactoryProvider = aggregateManagerWrapperFactoryProvider;
        this.geniUserProvider = geniUserProvider;
        this.sfaModelProvider = sfaModelProvider;
        this.epcManagerProvider = epcManagerProvider;
        this.proxySocketFactoryProvider = proxySocketFactoryProvider;
    }

    public UpdateUntilFinalOpStateTaskState createGetStatusUntilReadyOrFailTaskState(
            @Nonnull Job<?> job,
            @Nullable Duration checkReadyInterval,
            @Nullable Duration waitForReadyDelay,
            @Nonnull SfaExperimentPart experimentPart,
            @Nonnull DesiredStatus desiredStatus) {
        return new UpdateUntilFinalOpStateTaskState(job, experimentPart,
                checkReadyInterval, waitForReadyDelay, hltfProvider.get(),
                desiredStatus);
    }

    public CreateUserSpecListState createCreateUserSpecListState(Experiment experiment) {
        return new CreateUserSpecListState(experiment, geniUserProvider, sfaModelProvider.get());
    }


    public AllocateSliversOnAuthorityState createAllocateSliversOnAuthorityState(
            AbstractJob<?> job, SfaExperimentPart experimentPart, List<UserSpec> userSpecs,
            @Nullable RequestRspecSource overridingRequestRspecSource) {
        return new AllocateSliversOnAuthorityState(job, experimentPart, userSpecs, overridingRequestRspecSource,
                hltfProvider.get(), aggregateManagerWrapperFactoryProvider.get());
    }

    public AllocateSliversOnAuthorityState createAllocateSliversOnAuthorityState(
            AbstractJob<?> job, SfaExperimentPart experimentPart, List<UserSpec> userSpecs) {
        return createAllocateSliversOnAuthorityState(job, experimentPart, userSpecs, null);
    }

    public ProvisionSliversOnAuthorityState createProvisionSliversOnAuthorityState(
            AbstractJob<?> job, SfaExperimentPart experimentPart, List<UserSpec> userSpecs) {
        return new ProvisionSliversOnAuthorityState(experimentPart, userSpecs, hltfProvider.get());
    }

    public StopExperimentAtAuthorityPartState createStopExperimentAtAuthorityPartState(Job<?> job, SfaExperimentPart experimentPart) {
        return new StopExperimentAtAuthorityPartState(job, experimentPart, hltfProvider.get());
    }

    public ShareLanAtAuthorityPartState createShareLanState(Job<?> job, SfaExperimentPart experimentPart, String linkClientId, String sharedLanName) {
        return new ShareLanAtAuthorityPartState(job, experimentPart, hltfProvider.get(), linkClientId, sharedLanName);
    }

    public DistributeSshKeysState createDistributeSshKeysState(SetupSoftwareExperimentJob job){
        return new DistributeSshKeysState(job);
    }
    public WaitForExecuteFinishedState createWaitForExecuteFinishedState(@Nonnull SetupSoftwareExperimentJob job){
        return new WaitForExecuteFinishedState(job);
    }
    public ExecuteAnsibleServicesState createExecuteAnsibleServicesState(@Nonnull SetupSoftwareExperimentJob job){
        return new ExecuteAnsibleServicesState(job, geniUserProvider);
    }
    public List<State> createOrderedExperimentServicesStates(@Nonnull SetupSoftwareExperimentJob job){
        assert job.getExperiment().getExperimentSpecification() != null;
        assert job.getExperiment().getExperimentSpecificationBundle() != null;

        if (job.getExperiment().getExperimentSpecification() == null)
            throw new IllegalStateException("Internal error, createOrderedExperimentServicesStates called for experiment without ExperimentSpecification");
        if (job.getExperiment().getExperimentSpecificationBundle() == null)
            throw new IllegalStateException("Internal error, createOrderedExperimentServicesStates called for experiment without ExperimentSpecification (bundle)");

        List<State> res = new ArrayList<>();

        RemotePathResolver remotePathResolver = new RemotePathResolver(job);

        if (job.isRunAll()) {
            ExperimentSpecificationServicesDirsAndUploadsState dirAndUpload =
                    new ExperimentSpecificationServicesDirsAndUploadsState(job, remotePathResolver, geniUserProvider.getLoggedInGeniUser());
            res.add(dirAndUpload);
        }

        //if startPos is null, start from the beginning
        boolean mustRun = job.getStartPosition() == null;

        for (ExecuteSpec executeSpec : job.getExperiment().getExperimentSpecification().getExecutes()) {
            if (job.getStartPosition() != null && executeSpec == job.getStartPosition())
                mustRun = true;
            if (mustRun) {
                ExperimentSpecificationServicesSingleExecuteState singleExecuteState =
                        new ExperimentSpecificationServicesSingleExecuteState(
                                job,
                                remotePathResolver,
                                executeSpec,
                                false);
                res.add(singleExecuteState);
            }
            if (job.getStopPosition() != null && executeSpec == job.getStopPosition())
                mustRun = false;
        }
        if (job.getExperiment().getExperimentSpecification().getAnsible() != null) {
            for (ExecuteSpec executeSpec : job.getExperiment().getExperimentSpecification().getAnsible().getAnsibleHostSpec().getExecuteLists()) {
                if (job.getStartPosition() != null && executeSpec == job.getStartPosition())
                    mustRun = true;
                if (mustRun) {
                    ExperimentSpecificationServicesSingleExecuteState singleExecuteState =
                            new ExperimentSpecificationServicesSingleExecuteState(
                                    job,
                                    remotePathResolver,
                                    executeSpec,
                                    true);
                    res.add(singleExecuteState);
                }
                if (job.getStopPosition() != null && executeSpec == job.getStopPosition())
                    mustRun = false;
            }

            AnsibleSpec ansibleSpec = job.getExperiment().getExperimentSpecification().getAnsible();
            if (job.getStartPosition() != null && ansibleSpec == job.getStartPosition())
                mustRun = true;
            if (mustRun) {
                ExperimentSpecificationServicesAnsibleState ansibleState =
                        new ExperimentSpecificationServicesAnsibleState(
                                job,
                                geniUserProvider,
                                remotePathResolver,
                                ansibleSpec);
                res.add(ansibleState);
            }
            if (job.getStopPosition() != null && ansibleSpec == job.getStopPosition())
                mustRun = false;
        }

        //after all executes, make sure the SSH connections are closed
        res.add(new AssureAllSshClosedStateSlice(job));
        return res;
    }


    public ExperimentPartsSlicedState<ExperimentPartStateSlice> createUpdateSshKeysState(Job<?> job, Collection<UserSpec> userSpecs) {
        return new ExperimentPartsSlicedState<ExperimentPartStateSlice>("Update SSH keys on authorities", job.getExperiment(), epcManagerProvider.get()) {
            @Override
            public <EP extends ExperimentPart> ExperimentPartStateSlice createJobSlice(ExperimentPartController<EP> controller, EP experimentPart) {
                return controller.editSshKeys(job, experimentPart, userSpecs);
            }
        };
    }

//    public ExperimentPartsSlicedState<ExperimentPartStateSlice> createStopExperimentState(Job<?> job) {
//        return new StopExperimentState(job);
//    }

    public class StopExperimentState extends ExperimentPartsSlicedState<ExperimentPartStateSlice> {
        private final Job<?> job;

        private StopExperimentState(Job<?> job) {
            super("Delete all resources", job.getExperiment(), JobStateFactory.this.epcManagerProvider.get());
            this.job = job;
        }

        @Override
        public <EP extends ExperimentPart> ExperimentPartStateSlice createJobSlice(ExperimentPartController<EP> controller, EP experimentPart) {
            return controller.stop(job, experimentPart);
        }
    }
}
